@@ -1,7 +1,7 @@
 darcs_init () {
     WD=$1
     mkdir -p "$WD"
-    (cd "$WD" && darcs init)
+    (cd "$WD" && darcs init && echo "Chester McTester" >_darcs/prefs/author)
 
     export VC_DIFF="darcs diff"
     export VC_RM="true"
